/*
 * Copyright (C) 1996-2001 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include "config.h"

#include <stdio.h>
#include <rpc/rpc.h>
#include <errno.h>
#include <termios.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#define  GDHE_CLIENT_VAR
#include "gdhe/GDHE.h"
#include "GDHE_xdr_prot.h"

struct READER_REQUEST {
    char *name;
    int pause;
} pauseTab[] = {
    {"DISCONNECT", 0},
    {"PLACE_ROBOT", 0},
    {"SET_PLATFORM", 0},
    {"PLACE_DIAMOND", 0},
    {"PLACE_DOORS", 0},
    {"PLACE_WALLS", 0},
    {"PLACE_WALLS_REL", 0},
    {"PLACE_WALLS_ACC", 0},
    {"PLACE_FREE_SPACE", 0},
    {"PLACE_TRAJECTORY", 0},
    {"PLACE_PATH", 0},
    {"ERASE_ROBOT", 0},
    {"ERASE_DIAMOND", 0},
    {"ERASE_DOORS", 0},
    {"ERASE_WALLS", 0},
    {"ERASE_FREE_SPACE", 0},
    {"ERASE_TRAJECTORY", 0},
    {"ERASE_PATH", 0},
    {"PLACE_GOAL_O", 0},
    {"PLACE_GOAL_NO", 0},
    {"ERASE_GOAL", 0},
 
    {"EXEC_TRAJ", 0},
    {"MOVE_PLATFORM", 0},
 
    {"VIDEO_ON", 0},
    {"VIDEO_OFF", 0},
    {"LASER_ON", 0},
    {"LASER_OFF", 0},
 
    {"SELECT_ROBOT_TYPE", 0},
    {"UPDATE_PER_OBST", 0},
    {"ERASE_PER_OBST", 0},
 
    {"PLACE_ROBOT_TRAJ", 0},
    {"ERASE_ROBOT_TRAJ", 0},
    {"PLACE_ROBOT_PATH", 0},
    {"ERASE_ROBOT_PATH", 0},
 
    {"PLACE_RS_TRAJ", 0},
    {"ERASE_RS_TRAJ", 0},
 
    {"PLACE_DEPENDENCY_EDGE", 0},
    {"ERASE_DEPENDENCY_EDGE", 0},
    {"PLACE_DEPENDENCY_CYCLE", 0},
    {"ERASE_DEPENDENCY_CYCLE", 0},
 
    {"PLACE_MISSION_GOAL", 0},
    {"ERASE_MISSION_GOAL", 0},
    {NULL, 0}
};

int
setPause(char *event)
{
    int i;

    i = 0;
    while (pauseTab[i].name != NULL) {
	if (strcmp(pauseTab[i].name, event) == 0) {
	    pauseTab[i].pause = 1;
	    break;
	}
	i++;
    } /* while */
    if (pauseTab[i].name == NULL) {
	return(-1);
    } else {
	return(0);
    }
}

struct termios otty;
void 
save_tty(void)
{
    tcgetattr(fileno(stdin), &otty);
}

void 
restore_tty(void)
{
    tcsetattr(fileno(stdin), TCSADRAIN, &otty);
}    
    
int 
attend_touche(void)
{
    char buf;
    struct termios tty;
 
     /* se mettre en mode brut et noecho */
    memcpy(&tty, &otty, sizeof(struct termios));                 
    tty.c_lflag  &= ~(ICANON|ECHO);            /* mode non -canonique */
    tty.c_cc[VMIN] = 1;                 /* read 1 char */
    tty.c_cc[VTIME] = 0;                /* block read */
    tcsetattr(fileno(stdin), TCSAFLUSH, &tty);
    
    /* Lire un caractere */
    read(fileno(stdin), &buf, 1);
    return(buf);
}
 

int
kbhit(void)
{
    int n;
    char buf;
    struct termios tty;
 
    /* se mettre en mode brut et noecho */
    memcpy(&tty, &otty, sizeof(struct termios)); 
    tty.c_lflag  &= ~(ICANON|ECHO);    /* mode non -canonique */
    tty.c_cc[VMIN] = 0;  
    tty.c_cc[VTIME] = 0;
    tcsetattr(fileno(stdin), TCSANOW, &tty);

    n = read(fileno(stdin), &buf, 1);
    if (n == 0) {
        buf = -1;
    }
    return(buf);
}

/* 
 * operations sur les structs timeval avec retenues
 */
void
time_add(struct timeval *r, struct timeval *a, struct timeval *b)
{
    long retenue = 0;

    /* r = a + b */
    r->tv_usec = a->tv_usec + b->tv_usec;
    while (r->tv_usec > 1000000L) {
	r->tv_usec -= 1000000L;
	retenue++;
    }
    r->tv_sec = a->tv_sec + b->tv_sec + retenue;
}

void
time_difference(struct timeval *r, struct timeval *a, struct timeval *b)
{
    long retenue = 0;

    /* r = a - b */
    r->tv_usec = a->tv_usec - b->tv_usec;
    while (r->tv_usec < 0) {
	r->tv_usec += 1000000L;
	retenue++;
    }
    r->tv_sec = a->tv_sec - b->tv_sec - retenue;
}

void
time_multiply(struct timeval *r, struct timeval *a, int x)
{
    long retenue = 0;
    
    /* r = a * x */
    r->tv_usec = a->tv_usec * x;
    while (r->tv_usec > 1000000L) {
	r->tv_usec -= 1000000L;
	retenue++;
    }
    r->tv_sec = a->tv_sec * x + retenue;
}

void
time_divide(struct timeval *r, struct timeval *a, int x)
{
    
    long reste;

    /* r = a / x */
    r->tv_sec = a->tv_sec / x;
    reste = (a->tv_sec % x) * 1000000L;
    r->tv_usec = (a->tv_usec + reste) / x;
}
   
/*
 * Syncronisation avec les dates stocke'es dans le log
 */
void
synchro(struct timeval *date, int speed)
{
    static int first_time = 1;
    static struct timeval t0, offset;
    struct timeval now, pause;
    
    gettimeofday(&now, NULL);
    if (first_time) {
	/* Memorise la date de debut du log */
	memcpy(&t0, date, sizeof(struct timeval));

	/* Calcul de l'ecart entre la date du log et maintenant */
	time_difference(&offset, &now, date);
	first_time = 0;
	return;
    }
    if (speed != 1) {
	/* Recalcule la date du packet en fonction de l'accelaration */
	/* date = t0 + (date - t0)/speed */
	time_difference(date, date, &t0);
	time_divide(date, date, speed);
	time_add(date, &t0, date);
    }

    /* calcul delai d'attente */
    time_add(date, date, &offset);/* date += offset */
    time_difference(&pause, date, &now); /* pause = date - now */
    /* attente */
    if (pause.tv_sec >= 0 && pause.tv_usec > 0) {
	select(0, NULL, NULL, NULL, &pause);
    }
}

/*----------------------------------------------------------------------*/

int
main(int argc, char *argv[])
{
    char *host = NULL;
    char *filename;
    char *mpeg = NULL;
    FILE *input = NULL;
    int requests = 1;
    int millis = 20;
    int c, errflag = 0;
    XDR input_xdr, connected_xdr;
    int i, error, fini, counter;
    packet data;
    int compat = FALSE;
    int speed = 1;
    struct timeval packet_date;

    while ((c = getopt(argc, argv, "cd:f:h:lm:n:p:s:")) != EOF) {
	switch (c) {
	  case 'c':
	    compat = TRUE;
	    break;
	  case 'd':
	    millis = atoi(optarg);
	    break;
	  case 'f':
	    input = fopen(optarg, "r");
	    if (input == NULL) {
		perror(optarg);
		exit(2);
	    }
	    filename = optarg;
	    break;
	  case 'h':
	    host = optarg;
	    break;
	  case 'l':
	    fprintf(stderr, "evenements:\n");
	    for (i = 0; pauseTab[i].name != NULL; i++) {
		fprintf(stderr, "\t%s\n", pauseTab[i].name);
	    }
	    exit(0);
	    break;
	  case 'm':
	    mpeg = optarg;
	    break;
	  case 'n':
	    requests = atoi(optarg);
	    break;
	  case 'p':
	    if (setPause(optarg) == -1) {
		fprintf(stderr, "evenement inconnu %s\n", optarg);
		exit(2);
	    }
	    break;
	  case 's':
	    speed = atoi(optarg);
	    if (speed < 0) {
		fprintf(stderr, "speed doit etre >= 0\n");
		exit(2);
	    }
	    break;
	  case '?':
	    errflag++;
	} /* switch */
    } /* while */
    if (errflag) {
	fprintf(stderr, "usage: %s [-c][-f file][-h host][-n packets]"
		"[-d milli][-p event][-l][-s speed factor]\n", argv[0]);
	exit(2);
    }
    if (input == NULL) {	
	fprintf(stderr, "%s: need to specify a file name with -f\n", argv[0]);
	exit(2);
    }
    if (host == NULL) {
	host = "localhost";
    }
    errno = 0;
    xdrstdio_create(&input_xdr, input, XDR_DECODE);
    if (errno) {
	perror("xdrstdio_create");
	exit(1);
    }
    error = get_connection(host);
    if (error == 0) {	
	fprintf(stderr, "error: could not get connection to %s\n", host);
	exit(1);
    }
    connected_xdr = get_connected_xdr();
    
    /* C'est parti */
    fini = 0;
    counter = 0;
    save_tty();
    if (mpeg) {
	char buf[256];
	eval_expression("film_window");
	snprintf(buf, sizeof(buf), ".film.ogl mpegCaptureInit %s", mpeg);
	eval_expression(buf);
    }
    while (!fini) {
	data.args.pos = NULL;
	if (!compat) {
	    /* Lecture de la date du paquet */
	    if (xdr_timeval(&input_xdr, &packet_date) == FALSE) {
		fini = 1;
		break;
	    }
	    if (speed != 0 && !mpeg) {
		/* attente du delai */
		synchro(&packet_date, speed);
	    }
	}
	
	if (xdr_packet(&input_xdr, &data) == FALSE) {
	    fini = 1;
	    break;
	}
	if (xdr_packet(&connected_xdr, &data) == FALSE) {
	    fini = 1;
	}
	if (mpeg) {
	    /* Enregistrement de l'image */
	    eval_expression(".film.ogl redraw t");
	    eval_expression(".film.ogl mpegCaptureFrame");
	}
	if (pauseTab[data.command].pause != 0) {
	    printf("-- %s -- Press any key to continue...", 
		   pauseTab[data.command].name);
	    fflush(stdout);
	    if (attend_touche() == 'q') {
		fini = 1;
	    }
	    putchar('\n');
	}
	xdr_free((xdrproc_t)xdr_packet, (char *)&data);
	counter++;
	if (counter >= requests && millis > 0) {
	    /* tempo */
            struct timeval time_out;
            time_out.tv_usec = ( millis * 1000 ) % 1000000;
            time_out.tv_sec = millis / 1000;
            select ( 0, NULL, NULL, NULL, &time_out );
	    counter = 0;
	}
	c = kbhit();
	if (c == 'q' || c == 'Q') {
	    fini = 1;
	} else 	if (c > 0) {
	    printf("-- Pause -- Press any key to continue...");
	    fflush(stdout);
	    if (attend_touche() == 'q') {
		fini = 1;
	    }
	    putchar('\n');
	}
	if (feof(input)) {
	    fini = 1;
	}
    } /* while */
    if (mpeg) {
	eval_expression("mpegCaptureEnd");
    }
    data.command = DISCONNECT;
    xdr_packet(&connected_xdr, &data);
    xdr_destroy(&input_xdr);
    printf("Bye...\n");
    fclose(input);
    restore_tty();
    return 0;
} /* main */
	
